#!/usr/bin/env python

from cTriangle import *

def poly2tri(p, t=TriData()):
    """Convert Polygon to TriData structure"""
    import Polygon
    cTri = lambda a,b,c: ((a[0]+b[0]+c[0])/3.0, (a[1]+b[1]+c[1])/3.0)
    for j in range(len(p)):
        c = p[j]
        # append point
        if not c[0] in t.points:
            t.points.append(c[0])
        # append segment
        for i in range(1, len(c)):
            if not c[i] in t.points:
                t.points.append(c[i])
            t.segments.append((t.points.index(c[i-1]), t.points.index(c[i])))
            t.segmentMarkers.append(0)
        # close contour
        if c[0] != c[-1]:
            t.segments.append((t.points.index(c[-1]), t.points.index(c[0])))
            t.segmentMarkers.append(0)
        # if contour is hole, append hole point as center of first triangle
        # calculated by TriStrip
        if p.isHole(j):
            h = Polygon.TriStrip(Polygon.Polygon(c))
            if h and h[0]:
                a, b, c  = h[0][0:3]
                t.holes.append(((a[0]+b[0]+c[0])/3.0, (a[1]+b[1]+c[1])/3.0))
    return t

def plotSegments(filename, *tri):
    f = open(filename, 'w')
    for t in tri:
        for s in t.segments:
            x = tuple(t.points[s[0]]+t.points[s[1]])
            f.write('%g %g\n%g %g\n\n' % x)
    f.close()

def plotTriangles(filename, *tri):
    f = open(filename, 'w')
    for t in tri:
        for a in t.triangles:
            x = tuple(t.points[a[0]]+t.points[a[1]]+t.points[a[2]]+t.points[a[0]])
            f.write('%g %g\n%g %g\n%g %g\n%g %g\n\n' % x)
    f.close()


if __name__ == '__main__':
    import Polygon.Shapes
    p = Polygon.Shapes.Star() - Polygon.Shapes.Star(0.3, (0.2, 0.2), 10, 0.2)
    t = poly2tri(p)
    plotSegments('foo.gp', t)
    x = t.triangulate('QzpLAeq30')
    plotTriangles('bar.gp', x)
