#include <Python.h>
#include "triangle.h"
#include <string.h>

#ifndef TRIANGLEVERSION
#define TRIANGLEVERSION "UNKNOWN"
#endif


/* TriData */
typedef struct {
    PyObject_HEAD
    PyObject *points;
    PyObject *pointAttributes;
    PyObject *pointMarkers;
    PyObject *holes;
    PyObject *segments;
    PyObject *segmentMarkers;
    PyObject *triangles;
    PyObject *triangleAttributes;
    PyObject *triangleAreas;
    PyObject *neighbors;
    PyObject *regions;
    PyObject *edges;
    PyObject *edgeMarkers;
    PyObject *normals;
} TriData;

/* triangulate section*/

typedef struct triangulateio TrIO;

TrIO *newTrIO(void) {
    TrIO * t = (TrIO *)malloc(sizeof(TrIO));
    t->pointlist                    = NULL;
    t->pointattributelist           = NULL;
    t->pointmarkerlist              = NULL;
    t->numberofpoints               = 0;
    t->numberofpointattributes      = 0;
    t->trianglelist                 = NULL;
    t->triangleattributelist        = NULL;
    t->trianglearealist             = NULL;
    t->neighborlist                 = NULL;
    t->numberoftriangles            = 0;
    t->numberofcorners              = 3;
    t->numberoftriangleattributes   = 0;
    t->segmentlist                  = NULL;
    t->segmentmarkerlist            = NULL;
    t->numberofsegments             = 0;
    t->holelist                     = NULL;
    t->numberofholes                = 0;
    t->regionlist                   = NULL;
    t->numberofregions              = 0;
    t->edgelist                     = NULL;
    t->edgemarkerlist               = NULL;
    t->normlist                     = NULL;
    t->numberofedges                = 0;
    return t;
}

void deleteTrIO(TrIO *t, int borrowed) {
    if (t->pointlist              != NULL) free(t->pointlist);
    if (t->pointattributelist     != NULL) free(t->pointattributelist);
    if (t->pointmarkerlist        != NULL) free(t->pointmarkerlist);
    if (t->trianglelist           != NULL) free(t->trianglelist);
    if (t->triangleattributelist  != NULL) free(t->triangleattributelist);
    if (t->trianglearealist       != NULL) free(t->trianglearealist);
    if (t->neighborlist           != NULL) free(t->neighborlist);
    if (t->segmentlist            != NULL) free(t->segmentlist);
    if (t->segmentmarkerlist      != NULL) free(t->segmentmarkerlist);
    if (t->edgelist               != NULL) free(t->edgelist);
    if (t->edgemarkerlist         != NULL) free(t->edgemarkerlist);
    if (t->normlist               != NULL) free(t->normlist);
    if (!borrowed) {
	if (t->holelist           != NULL) free(t->holelist);
	if (t->regionlist         != NULL) free(t->regionlist);
    }
    free(t);
}

/* conversation section */
void dbl2list(PyObject *list, double *flat, int len, int n) {
    int i, j;
    PyObject *elem;
    while (PySequence_Size(list) > 0)
	PySequence_DelItem(list, 0);
    if (flat == NULL)
	return;
    for (i = 0; i < len; i++) {
        elem = PyTuple_New(n);
        for (j = 0; j < n; j++)
            PyTuple_SetItem(elem, j, PyFloat_FromDouble(flat[n*i+j]));
        PyList_Append(list, elem);
    }
}

void int2list(PyObject *list, int *flat, int len, int n) {
    int i, j;
    PyObject *elem;
    while (PySequence_Size(list) > 0)
	PySequence_DelItem(list, 0);
    if (flat == NULL)
	return;
    for (i = 0; i < len; i++) {
        elem = PyTuple_New(n);
        for (j = 0; j < n; j++)
            PyTuple_SetItem(elem, j, PyInt_FromLong((long)(flat[n*i+j])));
        PyList_Append(list, elem);
    }
}

double *list2dbl(PyObject *list, int *size, int n) {
    double *flat;
    int i, j, imax = PyList_Size(list);
    *size = imax; 
    if (imax == 0)
	return NULL;
    flat = (double *)malloc(sizeof(double)*imax*n);
    if (n == 1) {
	for (i=0; i<imax; i++)
	    flat[i] = PyFloat_AsDouble(PyNumber_Float(PySequence_Fast_GET_ITEM(list, i)));
    } else {
	PyObject *elem;
	for (i=0; i<imax; i++) {
	    elem = PySequence_Fast(PyList_GetItem(list, i), "illegal sequence");
	    for (j = 0; j < n; j++)
		flat[n*i+j] = PyFloat_AsDouble(PyNumber_Float(PySequence_Fast_GET_ITEM(elem, j)));
	}
    }  
    return flat;
}

int *list2int(PyObject *list, int *size, int n) {
    int *flat;
    int i, j, imax = PyList_Size(list);
    *size = imax;
    if (imax == 0)
	return NULL;
    flat = (int *)malloc(sizeof(int)*imax*n);
    if (n == 1) {
	for (i=0; i<imax; i++) 
	    flat[i] = (int)PyInt_AsLong(PyNumber_Int(PySequence_Fast_GET_ITEM(list, i)));
    } else {
	PyObject *elem;
	for (i=0; i<imax; i++) {
	    elem = PySequence_Fast(PyList_GetItem(list, i), "illegal sequence");
	    for (j = 0; j < n; j++)
		flat[n*i+j] = (int)PyInt_AsLong(PyNumber_Int(PySequence_Fast_GET_ITEM(elem, j)));
	}
    }
    return flat;
}

int checkArray(PyObject *list, int *m, int *n) {
    int i;
    PyObject *foo;
    if ((*m = PySequence_Size(list)) > 0) {
	foo = PyList_GetItem(list, 0);
	if (! PySequence_Check(foo))
	    *n = PySequence_Size(foo);
	else
	    return 0;
	for (i = 1; i < *m; i++) {
	    foo = PyList_GetItem(list, i);
	    if (! PySequence_Check(list))
		return 0;
	    if (PySequence_Size(foo) != *n)
		return 0;
        }
        return 1;
    }
    return 0;
}

/* conversation section */
int tridata2trio(TriData *d, TrIO *io) {
    int foo;
    PyObject *bar;
    io->pointlist               = list2dbl(d->points, &(io->numberofpoints), 2);
    io->pointmarkerlist         = list2int(d->pointMarkers, &foo, 1);
    if (PySequence_Size(d->pointAttributes) > 0) {
	bar = PySequence_GetItem(d->pointAttributes, 0);
	if (PySequence_Check(bar)) {
	    io->numberofpointattributes = PySequence_Size(bar);
	    io->pointattributelist = list2dbl(d->pointAttributes, &foo, 
					      io->numberofpointattributes);
	}
    }
    io->holelist            = list2dbl(d->holes, &(io->numberofholes), 2);
    io->segmentlist         = list2int(d->segments, &(io->numberofsegments), 2);
    io->segmentmarkerlist   = list2int(d->segmentMarkers, &foo, 1);
    io->trianglelist        = list2int(d->triangles, &(io->numberoftriangles), 3);
    if (PySequence_Size(d->triangleAttributes) > 0) {
	bar = PySequence_GetItem(d->triangleAttributes, 0);
	if (PySequence_Check(bar)) {
	    io->numberoftriangleattributes = PySequence_Size(bar);
	    io->triangleattributelist = list2dbl(d->triangleAttributes, &foo, 
						 io->numberoftriangleattributes);
	}
    }
    io->trianglearealist     = list2dbl(d->triangleAreas, &foo, 1);
    io->regionlist           = list2dbl(d->regions, &(io->numberofregions), 4);
    /* output only: neighbors, edges, edgeMarkers, normals*/
    return 1;
}

int trio2tridata(TrIO *io, TriData *d) {
    dbl2list(d->points, io->pointlist, io->numberofpoints, 2);
    int2list(d->pointMarkers, io->pointmarkerlist, io->numberofpoints, 1);
    dbl2list(d->pointAttributes, io->pointattributelist, io->numberofpoints, io->numberofpointattributes);
    dbl2list(d->holes, io->holelist, io->numberofholes, 2);
    int2list(d->segments, io->segmentlist, io->numberofsegments, 2);
    int2list(d->segmentMarkers, io->segmentmarkerlist, io->numberofsegments, 1);
    int2list(d->triangles, io->trianglelist, io->numberoftriangles, 3);
    dbl2list(d->triangleAttributes, io->triangleattributelist, io->numberoftriangles, io->numberoftriangleattributes);
    int2list(d->neighbors, io->neighborlist, io->numberoftriangles, 3);
    dbl2list(d->regions, io->regionlist, io->numberofregions, 4);
    int2list(d->edges, io->edgelist, io->numberofedges, 2);
    int2list(d->edgeMarkers, io->edgemarkerlist, io->numberofedges, 1);
    dbl2list(d->normals, io->normlist, io->numberofedges, 2);
    /* input only: triangleAreas, */
    return 1;
}


/* PYTHON section */

/* general Exception */
static PyObject *TriDataError;

/* standard methods */
PyObject * TriData_new(PyObject *self, PyObject *args); 
void       TriData_dealloc(PyObject *self);
PyObject * TriData_getAttr(PyObject *self, char * attrname);
PyObject * TriData_repr(PyObject *self);
/* my own */
PyObject * TriData_triangulate(PyObject *self, PyObject *args);
/* helpers */
PyObject * Triangle_Raise(PyObject *, char *);

/* some common exceptions ... */
#define ERR_ARG TriDataError, "Wrong number or type of arguments"
#define ERR_INV TriDataError, "Invalid argument"

/* raise an exception and return NULL */
PyObject * Triangle_Raise(PyObject *e, char *msg) {
    PyErr_SetString(e, msg);
    return NULL;
}

static PyMethodDef TriData_Methods[] = {
  {"triangulate", TriData_triangulate, METH_VARARGS},
  {NULL, NULL}, /* sentinel */
};

static PyTypeObject TriData_Type = {
#ifdef MS_WIN32    
  PyObject_HEAD_INIT(NULL)
#else
  PyObject_HEAD_INIT(&PyType_Type)
#endif  
  0,                               /*ob_size*/
  "TriData",                       /*tp_name*/
  sizeof(TriData),                 /*tp_basicsize*/
  0,                               /*tp_itemsize*/
  /* methods */
  (destructor)TriData_dealloc,     /*tp_dealloc*/
  0,                               /*tp_print*/
  (getattrfunc)TriData_getAttr,    /*tp_getattr*/
  0,                               /*tp_setattr*/
  0,                               /*tp_compare*/
  TriData_repr,                    /* (reprfunc)tp_repr*/
  0,                               /*tp_as_number*/
  0,                               /*tp_as_sequence*/
  0,                               /*tp_as_mapping*/
  0,                               /*tp_hash*/
};

PyObject * TriData_new(PyObject *self, PyObject *args) {
    TriData * t  = (TriData *)PyObject_NEW(TriData, &TriData_Type);
    t->points              = PyList_New(0);
    t->pointAttributes     = PyList_New(0);
    t->pointMarkers        = PyList_New(0);
    t->holes               = PyList_New(0);
    t->segments            = PyList_New(0);
    t->segmentMarkers      = PyList_New(0);
    t->triangles           = PyList_New(0);
    t->triangleAttributes  = PyList_New(0);
    t->triangleAreas       = PyList_New(0);
    t->neighbors           = PyList_New(0);
    t->regions             = PyList_New(0);
    t->edges               = PyList_New(0);
    t->edgeMarkers         = PyList_New(0);
    t->normals             = PyList_New(0);
    Py_INCREF(t->points);
    Py_INCREF(t->pointAttributes);
    Py_INCREF(t->pointMarkers);
    Py_INCREF(t->holes);
    Py_INCREF(t->segments);
    Py_INCREF(t->segmentMarkers);
    Py_INCREF(t->triangles);   
    Py_INCREF(t->triangleAttributes);
    Py_INCREF(t->triangleAreas);
    Py_INCREF(t->neighbors);
    Py_INCREF(t->regions);
    Py_INCREF(t->edges);
    Py_INCREF(t->edgeMarkers);
    Py_INCREF(t->normals);
    return (PyObject *)t;
}

void TriData_dealloc(PyObject *self) {
    Py_XDECREF(((TriData *)self)->points);
    Py_XDECREF(((TriData *)self)->pointAttributes);
    Py_XDECREF(((TriData *)self)->pointMarkers);
    Py_XDECREF(((TriData *)self)->holes);
    Py_XDECREF(((TriData *)self)->segments);
    Py_XDECREF(((TriData *)self)->segmentMarkers);
    Py_XDECREF(((TriData *)self)->triangles);
    Py_XDECREF(((TriData *)self)->triangleAttributes);
    Py_XDECREF(((TriData *)self)->triangleAreas);
    Py_XDECREF(((TriData *)self)->neighbors);
    Py_XDECREF(((TriData *)self)->regions);
    Py_XDECREF(((TriData *)self)->edges);
    Py_XDECREF(((TriData *)self)->edgeMarkers);
    Py_XDECREF(((TriData *)self)->normals);
    PyObject_Del(self);
}

PyObject * TriData_triangulate(PyObject *self, PyObject *args) {
    TrIO *in=newTrIO(), *out=newTrIO(), *vor=newTrIO();
    TriData *Out=(TriData *)TriData_new(Py_None, Py_None);
    char *options;
    if (! PyArg_ParseTuple(args, "s", &options))
        return Triangle_Raise(ERR_ARG);
    in = newTrIO();
    out = newTrIO();
    vor = newTrIO();
    tridata2trio((TriData *)self, in);
    triangulate(options, in, out, vor);
    trio2tridata(out, Out);
    deleteTrIO(in, 0);
    deleteTrIO(out, 1);
    deleteTrIO(vor, 1);
    return (PyObject *)Out;
}

PyObject * TriData_getAttr(PyObject *self, char *attrname) {
    PyObject *o=NULL;
    if (strcmp(attrname, "points") == 0)
	o = (PyObject *)(((TriData *)self)->points);
    else if (strcmp(attrname, "pointAttributes") == 0)
	o = (PyObject *)(((TriData *)self)->pointAttributes);
    else if (strcmp(attrname, "pointMarkers") == 0)
	o = (PyObject *)(((TriData *)self)->pointMarkers);
    else if (strcmp(attrname, "holes") == 0)
	o = (PyObject *)(((TriData *)self)->holes);
    else if (strcmp(attrname, "segments") == 0)
	o = (PyObject *)(((TriData *)self)->segments);
    else if (strcmp(attrname, "segmentMarkers") == 0)
	o = (PyObject *)(((TriData *)self)->segmentMarkers);
    else if (strcmp(attrname, "triangles") == 0)
	o = (PyObject *)(((TriData *)self)->triangles);
    else if (strcmp(attrname, "triangleAttributes") == 0)
	o = (PyObject *)(((TriData *)self)->triangleAttributes);
    else if (strcmp(attrname, "triangleAreas") == 0)
	o = (PyObject *)(((TriData *)self)->triangleAreas);
    else if (strcmp(attrname, "neighbors") == 0)
	o = (PyObject *)(((TriData *)self)->neighbors);
    else if (strcmp(attrname, "regions") == 0)
	o = (PyObject *)(((TriData *)self)->regions);
    else if (strcmp(attrname, "edges") == 0)
	o = (PyObject *)(((TriData *)self)->edges);
    else if (strcmp(attrname, "edgeMarkers") == 0)
	o = (PyObject *)(((TriData *)self)->edgeMarkers);
    else if (strcmp(attrname, "normals") == 0)
	o = (PyObject *)(((TriData *)self)->normals);
    if (o) {
	Py_INCREF(o);
	return o;
    } else
	return Py_FindMethod(TriData_Methods, self, attrname);
}


PyObject * TriData_repr(PyObject *self) {
    //TriData * t = (TriData *)self;
    return (PyObject *)PyString_FromString("<TriData object>");
}

static PyMethodDef TriangleMethods[] = {
    {"TriData",   TriData_new,  METH_VARARGS},
    {NULL,      NULL}        /* Sentinel */
};

void initcTriangle(void) {
    PyObject *m, *d;
#ifdef MS_WIN32
    TriData_Type.ob_type = &PyType_Type;
#endif    
    m = Py_InitModule("cTriangle", TriangleMethods);
    d = PyModule_GetDict(m);
    TriDataError = PyErr_NewException("cTriangle.Error", NULL, NULL);
    PyDict_SetItemString(d, "Error", TriDataError);
    PyDict_SetItemString(d, "TriDataType", (PyObject *)(&TriData_Type));
    PyDict_SetItemString(d, "version", 
			 (PyObject *)PyString_FromString(TRIANGLEVERSION));
}
